(
// Assumes that all the sounds in the folder and any subfolder only contains .wav or
// .aiff audio files
~loadSoundsFolder = {
	|soundPath|

	// Define a function with in a function
	var createBuffers = {
		arg pathName, soundsDictionary;
		if (pathName.isFile, {
			pathName.postln;
			soundsDictionary.put(pathName.fileName, Buffer.read(s, pathName.fullPath));
		});
		if (pathName.isFolder, {
			var entries = pathName.entries, index = 0;
			while({index < entries.size}, {
				createBuffers.value(entries[index], soundsDictionary);
				index = index + 1;
			});
		});
	};
	var soundsFolderPath = PathName.new(soundPath);
	var soundsDictionary = Dictionary[];
	createBuffers.value(soundsFolderPath, soundsDictionary);
	soundsDictionary; // return the sounds Dictionary
};

~path = thisProcess.nowExecutingPath.dirname;
~soundsDict = ~loadSoundsFolder.(~path +/+ "sounds");

SynthDef(\play_stereo, {
	arg outBus = 0, amp = 1, buf, pan = 0;
	var sig;

	sig = PlayBuf.ar(2, buf, doneAction: 2) * amp;
	sig = Balance2.ar(sig[0], sig[1], pan);
	Out.ar(outBus, sig);
}).add;

SynthDef(\play_mono, {
	arg outBus = 0, amp = 1, buf, pan = 0;
	var sig;

	sig = PlayBuf.ar(1, buf, doneAction: 2) * amp;
	sig = Pan2.ar(sig, pan);
	Out.ar(outBus, sig);
}).add;

SynthDef(\fmShimmer, {
	arg outBus = 0, freq = 400, modIndex = 1, harmRatio = 1, pan = 0,
	atk = 0.05, dec = 0.05, susLvl = 0.8, rel = 1.5, curve = -4, amp = 0.6, indexFactor = 2, gate = 1;
	var sig, car, mod, freq_m, env, indexEnv, k_f;

	// Control amplitude of sidebands by ramping them up and then down
	// based on modIndex and indexFactor
	indexEnv = EnvGen.kr(Env.new(
		[
			modIndex, // Minimum
			modIndex * indexFactor, // Maximum
			modIndex // Minimum
		],
		[atk + dec, rel] // Times to reach maximum and then minimum
	));

	// Compute the frequency modulation
	freq_m = harmRatio * freq;
	k_f = indexEnv * freq_m;
	mod = SinOsc.ar(freq_m, 0, k_f);
	car = SinOsc.ar(freq + mod, 0, 1);

	// Envelope for the amplitude of the signal
	env = Env.adsr(atk, dec, susLvl, rel, amp, curve);
	env = EnvGen.kr(env, gate, doneAction: 2);

	sig = Pan2.ar(car * env, pan);

	Out.ar(outBus, sig);
}).add;

// This is a classic kind of subtractive synth modeled on the Moog with
// six chorused saw oscillators -> filter -> envelope -> output
SynthDef(\saw, {
	arg outBus = 0, freq = 440, pan = 0, amp = 0.3, gate = 1, cutoff_freq = 2000, q = 20,
	atkTime = 0.05, decTime = 0.05, susLvl = 0.8, relTime = 0.2;
	var numFilters = 3, sig = [0, 0], env, cutoff;

	numFilters.do({
		sig = sig + Saw.ar(
			freq: {Rand(freq * 0.99, freq * 1.01)} ! 2,
			mul: amp/numFilters
		);
	});

	env = EnvGen.kr(Env.adsr(
		attackTime: atkTime,
		decayTime: decTime,
		sustainLevel: susLvl,
		releaseTime: relTime
	), gate, doneAction: 2);

	cutoff = cutoff_freq * EnvGen.kr(Env([1, 2, 1], [1, 1]));

	sig = RLPF.ar(sig, cutoff, 1/q);
	sig = sig * env;

	sig = Balance2.ar(sig[0], sig[1], pan);
	Out.ar(outBus, sig);
}).add;

SynthDef(\reverb, {
	arg outBus = 0, inBus, mix = 0.4;
	var dry, wet, sig, numFilters = 8;

	dry = In.ar(inBus, 2);

	wet = dry;
	numFilters.do({
		wet = AllpassN.ar(
			in: wet,
			maxdelaytime: 0.2,
			delaytime: {Rand(0.05, 0.15)} ! 2,
			decaytime: {Rand(1, 2)} ! 2
		);
	});

	wet = LPF.ar(wet, 3500);
	sig = dry.blend(wet, mix);

	Out.ar(outBus, sig);
}).add;

SynthDef(\master, {
	arg outBus = 0, inBus, amp = 1;
	var sig;

	sig = In.ar(inBus, 2);

	// Any effects you might like to the overall track heere
	sig = sig * amp;

	Out.ar(outBus, sig);
}).add;
)

// Drum Sounds
Synth(\play_stereo, [\buf, ~soundsDict["Clank10.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["Clap10.wav"]]);
Synth(\play_mono, [\buf, ~soundsDict["CH.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["bass.wav"]]);
Synth(\play_mono, [\buf, ~soundsDict["subbass.wav"]]);
Synth(\play_mono, [\buf, ~soundsDict["snare.wav"]]);

// Piano Sounds
Synth(\play_stereo, [\buf, ~soundsDict["Piano.ff.Bb5.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["Piano.ff.C5.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["Piano.ff.C6.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["Piano.ff.G5.wav"]]);
Synth(\play_stereo, [\buf, ~soundsDict["Piano.mf.C5.wav"]]); // note the mf (mezzoforte) is softer
Synth(\play_stereo, [\buf, ~soundsDict["Piano.mf.Bb4.wav"]]);

/*********
* PIECE *
*********/

// Opening Chords

(
Pbindef.clear;
~clock = TempoClock(54/60);
~reverbBus = Bus.audio(s, 2);
~masterBus = Bus.audio(s, 2);
~verb = Synth(\reverb, [\mix, 0.1, \inBus, ~reverbBus, \outBus, ~masterBus]);
~master = Synth(\master, [\amp, 0.9, \inBus, ~masterBus], ~verb, \addAfter);

Pbindef(
	\pad,
	\instrument, \fmShimmer,
	\midinote, Pseq([
		[60, 63, 65, 67],
		[60, 62, 65, 65]
	], inf),
	\pan, [-0.6, -0.2, 0.2, 0.6], // spread chords across stereo field
	\atk, 1,
	\dec, 1,
	\rel, 3,
	\legato, 0.6, // controls gate release
	\amp, 0.1,
	\curve, 2,
	\dur, 4,
	\out, ~reverbBus
).play(~clock, quant: 4);
)

// Basic Drum Pattern
// The general idea of the drum pattern is to make it better match our intuition
// of how we think about drums (i.e., as a 4x4 grid where each row is a beat
// subdivided into sixteenth notes).  You can create any number of keys and then
// reference them using Pfunc or Pkey.  Here if the grid has a 1 that means play
// the note and 0 means to rest.
(
Pbindef(
	\kick,
	\instrument, \play_stereo,
	\pat, Pseq([
		1, 0, 0, 0,
		0, 0, 0, 0,
		0, 0, 0, 0,
		0, 1, 0, 0
	], inf),
	\amp, Pfunc({|event| if(event[\pat] == 1, {1}, {Rest()})}),
	\buf, ~soundsDict["bass.wav"],
	\dur, 0.25,
	\outBus, ~masterBus
).play(~clock, quant: 4);

Pbindef(
	\subkick,
	\instrument, \play_mono,
	\pat, Pseq([
		0, 0, 0, 1,
		0, 0, 0, 0,
		0, 0, 0, 1,
		0, 0, 0, 0
	], inf),
	\amp, Pfunc({|event| if(event[\pat] == 1, {1}, {Rest()})}),
	\buf, ~soundsDict["subbass.wav"],
	\dur, 0.25,
	\outBus, ~masterBus
).play(~clock, quant: 4);

Pbindef(
	\snare,
	\instrument, \play_mono,
	\pat, Pseq([
		0, 0, 0, 0,
		1, 0, 0, 0,
		0, 0, 0, 0,
		1, 0, 0, 0
	], inf),
	\buf, ~soundsDict["snare.wav"],
	\amp, Pfunc({|event| if(event[\pat] == 1, {1}, {Rest()})}),
	\dur, 0.25,
	\outBus, ~masterBus
).play(~clock, quant: 4);
)

// Hi-Hat

(
Pbindef(
	\ch,
	\instrument, \play_mono,
	\buf, ~soundsDict["CH.wav"],
	\amp, Pseq([Rest(), 0.80, 0.9, 1], inf),
	\dur, 0.25,
	\outBus, ~masterBus
).play(~clock, quant: 4);
)

// Piano lick
// The sounds dictionary holds a dictionary of buffers where each key is the name
// of the audio file and the value is a buffer.  Cycle through the various buffers
// to play different piano notes.
(
~pianoMelody = Pseq([
	~soundsDict["Piano.ff.C5.wav"],
	~soundsDict["Piano.ff.Bb5.wav"],
	~soundsDict["Piano.ff.Bb5.wav"],
	~soundsDict["Piano.ff.C5.wav"],
	~soundsDict["Piano.ff.C6.wav"],
	~soundsDict["Piano.ff.C6.wav"],
	~soundsDict["Piano.ff.C5.wav"],
	~soundsDict["Piano.ff.Bb5.wav"],
	Rest(),
	~soundsDict["Piano.ff.C5.wav"],
	~soundsDict["Piano.ff.Bb5.wav"],
	~soundsDict["Piano.ff.Bb5.wav"],
	~soundsDict["Piano.ff.C5.wav"],
	~soundsDict["Piano.ff.G5.wav"],
	Rest(),
	~soundsDict["Piano.ff.C5.wav"],
], inf);

Pbindef(
	\piano,
	\instrument, \play_stereo,
	\buf, ~pianoMelody,
	\dur, 0.25,
	\amp, 1.2,
	\out, ~reverbBus
).play(~clock, quant: 8)
)

// FM Bass sound

(
Pbindef(
	\bass,
	\instrument, \fmShimmer,
	\midinote, Pseq([36, 36, 34, 34], inf),
	\amp, 0.3,
	\dur, Pseq([0.75, 3.25], inf),
	\modIndex, 2, // Keep mod index relatively low so we get a good bass sound
	\outBus, ~masterBus
).play(~clock, quant: 8);
)

/*************
 * Section 2 *
 *************/

(
// Here we need to ensure that already playing patterns
// get changed at the next measure instead of updated immediately.
// You can schedule events to happen at a later moment in time using
// the method .schedAbs
~clock.playNextBar({
	// Stop beats and bass
	Pbindef(\snare).stop();
	Pbindef(\kick).stop();
	Pbindef(\subkick).stop();
	Pbindef(\ch).stop();
	Pbindef(\bass).stop();
});

~pianoMelody2 = Pseq([
	Pseq([~soundsDict["Piano.mf.C5.wav"]], 4),
	Pseq([~soundsDict["Piano.mf.Bb4.wav"]], 4),
], inf);

// New piano lick
Pbindef(
	\piano,
	\buf, ~pianoMelody2,
	\dur, Pseq([0.75, 0.75, 2, 0.5], inf),
	\amp, Pseq([2, 2.5, 3, 2], inf),
).quant_(8); // Ensures that the change happens on a multiple of two measures
)

// Play the audio clip.  Notice we do not need to use a pattern here because
// we just want to play this clip once.
Synth(\play_stereo, [\buf, ~soundsDict["dont_let_him.wav"]])

/*************
 * Section 3 *
 *************/
// Make a recap but with a new synth lick to make it sound different and a little
// more climactic.

(
Pbindef(\snare).play(~clock, quant: 4);
Pbindef(\kick).play(~clock, quant: 4);
Pbindef(\subkick).play(~clock, quant: 4);
Pbindef(\ch).play(~clock, quant: 4);
Pbindef(\bass).play(~clock, quant: 4);

Pbindef(
	\piano,
	\buf, ~pianoMelody,
	\dur, 0.25,
	\amp, 1.2,
).quant_(8);

Pbindef(\saw_dream,
	\instrument, \saw,
	\midinote, Prand([60, 62, 65, 67], inf) + Prand([-12, 0, 12], inf),
	\legato, 1,
	\cutoff_freq, Pseg([2000, 3000, 2000], [7, 7], repeats: inf),
	\amp, Pseg([0.25, 0.35, 0.25], [10, 10], repeats: inf),
	\pan, Pwhite(-0.6, 0.6),
	\atkTime, 0.1,
	\dur, Pwrand([1/8, 3/8], [0.95, 0.05], inf)
).play(~clock, quant: 4)
)

// Change some parts of the saw pattern
// Execute in any order
Pbindef(\saw_dream, \cutoff_freq, Pseg([1000, 2000, 1000], [4, 4], repeats: inf)); // darker
Pbindef(\saw_dream, \cutoff_freq, Pseg([2000, 3000, 2000], [7, 7], repeats: inf)); // brighter
Pbindef(\saw_dream, \midinote, Prand([60, 62, 65, 67, 70], inf) + Prand([-12, 0, 12], inf));
Pbindef(\saw_dream, \midinote, Prand([60, 62, 65, 67], inf) + Prand([0, 12, 24], inf));


/**********
 * ENDING *
 **********/

// Just the chords and the piano and stop everything else
(
Pbindef(\pad, \chords, Pseq([[60, 63, 65, 67], [60, 62, 65, 65]], 2)).quant_(8);

~clock.playNextBar({
	// Stop beats and bass
	Pbindef(\snare).stop;
	Pbindef(\kick).stop;
	Pbindef(\subkick).stop;
	Pbindef(\ch).stop;
	Pbindef(\bass).stop;
	Pbindef(\saw_dream).stop;
});

Pbindef(
	\piano,
	\buf, ~pianoMelody2,
	\dur, Pseq([0.75, 0.75, 2, 0.5], inf),
	\amp, Pseq([
		Pseq([2, 2.5, 3, 2], 3), 4
	])
).quant_(8); // Ensures that the change happens on every two measures
)

/***********
 * CLEANUP *
 ***********/

(
Pbindef.clear;
~reverbBus.free;
~verb.free;
)